import java.beans.*;
import java.text.ParseException;
import java.util.List;
import java.util.ArrayList;
import java.util.ResourceBundle;
import java.util.regex.*;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.swing.table.AbstractTableModel;

import bsh.Interpreter;
import bsh.Primitive;
import bsh.NameSpace;
import bsh.EvalError;

import com.taco.data.ArrayUtilities;
import com.taco.data.INoReturnObservableMap;
import com.taco.text.CPunctuationState;
import com.taco.util.ChainedResourceBundleFactory;
import com.taco.swinger.SwingInvokeProxyFactory;
import com.taco.swinger.text2gui.*;

/** A demo application that evaluates expressions with BeanShell, and has a
 *  calculator-like interface.
 */
public class Calculator implements PropertyChangeListener {
    public Calculator() { }
    
    /** The table model for the variable / value table. It contains all
     *  variables in the interpreter.
     */
    protected class VariableTableModel extends AbstractTableModel {                
        public int getRowCount() {
            return _variableNames.size();
        }
        
        public int getColumnCount() { 
            return 3;
        }
        
        public Object getValueAt(int row, int column) {
            if (column == 0) {
                return _variableNames.get(row);   
            } else if (column == 1) {
                return _variableTypeNames.get(row);                 
            } else {
                return _variableValues.get(row);
            }                          
        }
         
        public String getColumnName(int column) {
            if (column == 0) {
                return _bundle.getString("variableText");
            } else if (column == 1) {
                return _bundle.getString("typeText");
            } else {
                return _bundle.getString("valueText");
            }
        }
        
        /** Update the table model in response to the interpreter changing
         *  state.
         */
        protected void _update() {
            NameSpace nameSpace = interpreter.getNameSpace();
            
            _variableNames.clear();
            _variableValues.clear();
            _variableTypeNames.clear();
            
            String[] varNames = nameSpace.getVariableNames();
            _variableNames.ensureCapacity(varNames.length);
            _variableTypeNames.ensureCapacity(varNames.length);
            _variableValues.ensureCapacity(varNames.length);
            
            ArrayUtilities.addElements(_variableNames, varNames);
            
            for (int i = 0; i < varNames.length; i++) {
                Object value = null;
                try {
                    value = nameSpace.getVariable(varNames[i]);                                                                                
                } catch (Exception e) {
                    _variableValues.add(null);
                    _variableTypeNames.add("Unknown");
                    continue;
                }
                
                _variableValues.add(value);
                                                
                if (value == null) {
                    _variableTypeNames.add("<nulltype>");
                } else if (value instanceof Primitive) {
                    Class type = ((Primitive) value).getType();
                    _variableTypeNames.add(type.getName());
                } else {                                       
                    _variableTypeNames.add(value.getClass().getName());
                }
            }

            fireTableDataChanged();                        
        }        
        
        protected ArrayList _variableNames = new ArrayList();
        protected ArrayList _variableTypeNames = new ArrayList();
        protected ArrayList _variableValues = new ArrayList();        
    }
    
    /** Process property changes in the argument map due to the GUI 
     *  changing.
     */
    public void propertyChange(PropertyChangeEvent event) {
        String propertyName = event.getPropertyName();        
        Object value = event.getNewValue();
        
        if ("input".equals(propertyName)) {
            System.out.println("Evaluating");
            final Object obj = _fixupAndEvaluate((String) value);            
            _argMap.putNoReturn("result", "" + obj);
            _varTableModel._update();                
        } else if ("degreeMode".equals(propertyName)) {            
            if (value instanceof Boolean) {
                _degreeMode = ((Boolean) value).booleanValue();
                System.out.println("degreeMode = " + _degreeMode);
            }            
        }                
    }

    /** Given an expression entered by the user, fix it before sending it
     *  to the BeanShell interpreter. If an error occurs, return "ERROR".
     */
    protected Object _fixupAndEvaluate(String expr) {                
        CharSequence cs = _addAssignment(expr);
        
        // Replace occurances of "sin" by "Math.sin".        
        cs = _fixMathFunctions(cs);
        
        cs = _fixMathConstants(cs); 
                
        cs = _fixDegrees(cs);
                                               
        expr = cs.toString();
                        
        System.out.println("Fixed expr = " + expr);
                                        
        try {
            return interpreter.eval(expr);
        } catch (EvalError ee) {
            return "ERROR";
        }                                 
    }
    
    /** If no variable is assigned, assign to "mem". */
    protected CharSequence _addAssignment(CharSequence cs) {
        Matcher m = _STARTS_WITH_ASSIGN_PATTERN.matcher(cs);
        
        if (m.lookingAt()) {
            return cs;
        } else {
            return "mem=" + cs.toString(); 
        }        
    }
    
    /** Convert function calls like <code>sin()</code> to
     *  <code>Math.sin()</code>.
     */
    protected CharSequence _fixMathFunctions(CharSequence cs) {
        Matcher m = _MATH_FUNCTIONS.matcher(cs);
        
        StringBuffer sb = new StringBuffer();
        
        while (m.find()) {
            m.appendReplacement(sb, "Math.$1(");
        }
        m.appendTail(sb);
        
        return sb;
    }
    
    /** Replace occurrances constants like "PI" with "Math.PI". */
    protected CharSequence _fixMathConstants(CharSequence cs) {
        
        Matcher m = _MATH_CONSTANTS.matcher(cs);
        
        StringBuffer sb = new StringBuffer();
        
        while (m.find()) {
            m.appendReplacement(sb, "Math." + m.group(1).toUpperCase());
        }
        m.appendTail(sb);
        
        return sb;
    }
    
    /** Change the expression so that angles are implicitly in degrees. */
    protected CharSequence _fixDegrees(CharSequence cs) {
        if (_degreeMode) {
            cs = _fixRadianArgumentFunctions(cs);
            cs = _fixRadianReturnFunctions(cs);    
        }         
        return cs;
    }        
    
    /** Change functions that take one radian argument to convert to 
     *  degrees first. 
     */
    protected CharSequence _fixRadianArgumentFunctions(CharSequence cs) {
        Matcher m = _RADIAN_ARGUMENT_FUNCTIONS.matcher(cs);
        
        StringBuffer sb = new StringBuffer();
        
        int startIndex = 0;
        while (m.find(startIndex)) {
            int openParenIndex = m.end() - 1;
            int closeParenIndex = -1;
            // Match the parenthesis
            try {
                closeParenIndex = new CPunctuationState().matchPunctuation(
                 cs, openParenIndex);
            } catch (ParseException pe) {
                return "ERROR";   
            }
            
            // Append everything up to the pattern.            
            sb.append(cs.subSequence(startIndex, m.start()));
                        
            sb.append("Math.").append(m.group(1)).append("(Math.toRadians(");
                        
            // recursively fix the stuff inside the parens                
            CharSequence arg = _fixRadianArgumentFunctions(
             cs.subSequence(openParenIndex + 1, closeParenIndex));
            
            if ("ERROR".equals(arg)) {
                return arg;
            }
              
            sb.append(arg.toString());   
            sb.append("))");                
                                   
            startIndex = closeParenIndex + 1;
        }
        
        sb.append(cs.subSequence(startIndex, cs.length()));        
                
        return sb;
    }

    /** Fix calls to functions that return angles in radians so that the
     *  returned value is converted to degrees. 
     */
    protected CharSequence _fixRadianReturnFunctions(CharSequence cs) {
        Matcher m = _RADIAN_RETURN_FUNCTIONS.matcher(cs);
        
        StringBuffer sb = new StringBuffer();
        
        int startIndex = 0;
        while (m.find(startIndex)) {
            int openParenIndex = m.end() - 1;
            int closeParenIndex = -1;
            // Match the parenthesis
            try {
                closeParenIndex = new CPunctuationState().matchPunctuation(
                 cs, openParenIndex);
            } catch (ParseException pe) {
                return "ERROR";   
            }
                                   
            // Append everything up to the pattern.                        
            sb.append(cs.subSequence(startIndex, m.start()));
                                    
            sb.append("Math.toDegrees(Math.").append(m.group(1)).append("(");
                                    
            // recursively fix the stuff inside the parens                
            CharSequence arg = _fixRadianReturnFunctions(
              cs.subSequence(openParenIndex + 1, closeParenIndex));

            if ("ERROR".equals(arg)) {
                return arg;
            }
              
            sb.append(arg.toString());   
            sb.append("))");                
                                   
            startIndex = closeParenIndex + 1;
        }
        
        sb.append(cs.subSequence(startIndex, cs.length()));        
                
        return sb;            
    }        
    
    public void initGUI() {
        _bundle = ChainedResourceBundleFactory.DEFAULT_INSTANCE.getBundle(
          getClass().getName() + "ResourceBundle");
          
        _argMap = SwingInvokeProxyFactory.makeSwingInvokeProxyMap();
        _argMap.putNoReturn("result", "0");
        _argMap.putNoReturn("varTableModel", _varTableModel);
        _argMap.putNoReturn("degreeMode", new Boolean(_degreeMode));
        
        _argMap.addPropertyChangeListener("input", this);
        _argMap.addPropertyChangeListener("degreeMode", this);
        
        JFrame frame = null;
        
        try {
            frame = (JFrame) DispatchingComponentConverter.DEFAULT_INSTANCE.
              toObject(_bundle, "mainFrame", _argMap, null);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
        
        frame.pack();
        frame.setVisible(true);          
    }
    
    public static void main(String[] args) {
        final Calculator c = new Calculator();
        
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                c.initGUI();
            }
        });
    }
             
    protected ResourceBundle _bundle = null;
    protected INoReturnObservableMap _argMap = null;
    protected Interpreter interpreter = new Interpreter();     
   
    protected VariableTableModel _varTableModel = new VariableTableModel();
    
    protected boolean _degreeMode = false;
    
    private static final Pattern _STARTS_WITH_ASSIGN_PATTERN = 
      Pattern.compile("\\w+\\s*(?:\\[.*?\\])?=", Pattern.DOTALL);
      
    private static final Pattern _MATH_FUNCTIONS =
      Pattern.compile("\\b(abs|acos|asin|atan|atan2|ceil|exp|floor|log|max|" +
        "min|pow|random|rint|round|sin|cos|tan|toDegrees|toRadians|sqrt)" +
        "\\s*\\(");
        
    private static final Pattern _MATH_CONSTANTS =
      Pattern.compile("\\b(E|PI)\\b", Pattern.CASE_INSENSITIVE);
      
    private static final Pattern _RADIAN_ARGUMENT_FUNCTIONS=
      Pattern.compile("\\bMath\\.(sin|cos|tan)\\s*\\(");
      
    private static final Pattern _RADIAN_RETURN_FUNCTIONS=
      Pattern.compile("\\bMath\\.(asin|acos|atan2?)\\s*\\(");      
}
